/*
 * SeqnoErrorModel: specify a list of TCP seqno's to drop
 * can be specified in any order
 */

#include "errmodel-split.h"

static class SeqnoErrorModelClass : public TclClass {
public:
	SeqnoErrorModelClass() : TclClass("ErrorModel/List/Seqno") {}
	TclObject* create(int, const char*const*) {
		return (new SeqnoErrorModel);
	}
} class_seqno_error_model;

SeqnoErrorModel::SeqnoErrorModel() : ListErrorModel(), flowid_(0) {
	unit_ == EU_PKT;
}

int SeqnoErrorModel::corrupt(Packet* p)
{
	/* assumes droplist_ is sorted */
	int rval = 0;	// no drop

	// Empty droplist, do not corrupt
	if (droplist_ == NULL || dropcnt_ == 0) {
		fprintf(stderr, "warning: SeqnoErrorModel: null drop list\n");
		return 0;
	}

	hdr_cmn *ch = hdr_cmn::access(p);
	hdr_ip  *iph = hdr_ip::access(p);
	int flowid = iph->flowid();

	// If the packet is not a TCP segment or does not belong
	// to the specified flow do not corrupt
	if((ch->ptype_ != PT_TCP)||(flowid != flowid_)) return 0;

	hdr_tcp *tcph = hdr_tcp::access(p);
	int seqno  = tcph->seqno();

	if ((cur_ < dropcnt_) && droplist_[cur_] == seqno) {
		rval = 1;
		cur_++;
	}
	cnt_++;

	return (rval);
}

int SeqnoErrorModel::command(int argc, const char*const* argv) {
	Tcl& tcl = Tcl::instance();
	if (strcmp(argv[1], "flowid") == 0) {
		flowid_ = atoi(argv[2]);
		printf("Errormodel flowid_: %d , argv: %s\n", flowid_, argv[2]);
		return(TCL_OK);
	}
	return (ListErrorModel::command(argc, argv));
}

