// Agent/TCP/St -- StTcpAgent

#include "tcp-st.h"
#include "../common/flags.h"
#include "../common/ip.h"


//
// OTCL linkage
//
//
//
//

static class StTcpAgentClass : public TclClass {
public:
	StTcpAgentClass() : TclClass("Agent/TCP/StTcpAgent") {}
	TclObject* create(int, const char*const*) {
		return (new StTcpAgent());
	}
} class_StTcp;

static class StTcpSinkClass : public TclClass {
public:
	StTcpSinkClass() : TclClass("Agent/TCPSink/DelAck/StTcpSink") {}
	TclObject* create(int, const char*const*) {
		return (new StTcpSink(new NAcker));
	}
} class_StTcpSink;

static class SplitTcpSinkClass : public TclClass {
public:
	SplitTcpSinkClass() : TclClass("Agent/TCPSink/DelAck/Split") {}
	TclObject* create(int, const char*const*) {
		return (new SplitTcpSink(new Acker));
	}
} class_SplitTcpSink;

static class SplitStTcpSinkClass : public TclClass {
public:
	SplitStTcpSinkClass() : TclClass("Agent/TCPSink/DelAck/St/Split") {}
	TclObject* create(int, const char*const*) {
		return (new SplitStTcpSink(new NAcker));
	}
} class_SplitStTcpSink;


//
// StTcpAgent implementation
//
//
//
//

StTcpAgent::StTcpAgent() :
	NewRenoTcpAgent(),
	highnack_(0),
	nack_(0),
	rtxcount_(0),
	rtx_allow_(0),
	rtx_stop_(-1),
	rtx_last_(-1),
	nrtx_(0),
	dupnack_(0),
	rwnd_(1),
	qwnd_(1.0),
	myrecover_(0),
	mydupwnd_(0)
{
	bind("highnack_", &highnack_);
	bind("nack_", &nack_);
	bind("rtxcount_", &rtxcount_);
	bind("rtx_allow_", &rtx_allow_);
	bind("rtx_stop_", &rtx_stop_);
	bind("rtx_last_", &rtx_last_);
	bind("nrtx_", &nrtx_);
	bind("dupnack_", &dupnack_);
	bind("rwnd_", &rwnd_);
	bind("qwnd_", &qwnd_);
	bind("myrecover_", &myrecover_);
	bind("mydupwnd_", &mydupwnd_);
}

int StTcpAgent::window() {
	int win = (int)(rwnd_) + int(mydupwnd_);
	return (win);
}

double StTcpAgent::windowd() {
	double win = (double)(rwnd_) + double(mydupwnd_);
	return (win);
}

//
// Overrides RenoTcpAgent::timeout() to set rwnd_ to 1
//
void StTcpAgent::timeout(int tno) {
	if (tno == TCP_TIMER_RTX) {
		rwnd_ = 1;
		mydupwnd_ = 0;
		myrecover_ = 0;
		highnack_ = highest_ack_;
		rtxcount_ = 0;
		rtx_allow_ = 0;
		rtx_stop_ = -1;
		nrtx_ = 0;
		RenoTcpAgent::timeout(tno);
	} else {
		timeout_nonrtx(tno);
	}
}

// Overrides NewRenoTcpAgent::recv() 
void StTcpAgent::recv(Packet *pkt, Handler*) {
	hdr_tcp *tcph = hdr_tcp::access(pkt);
#ifdef notdef
	if (pkt->type_ != PT_ACK) {
		Tcl::instance().evalf("%s error \"received non-ack\"",
				      name());
		Packet::free(pkt);
		return;
	}
#endif


//	printf("StTcpAgnet receive a packet");
	rwnd_ = tcph->tcp_flags_;

	++nackpack_;
	ts_peer_ = tcph->ts();
	int ecnecho = hdr_flags::access(pkt)->ecnecho();
	if (ecnecho && ecn_)
		ecn(tcph->seqno());
	recv_helper(pkt);
	if (tcph->seqno() > last_ack_) {
		if(tcph->seqno() >= myrecover_) {
		//new ack: ACK >= myrecover_
			mydupwnd_ = 0;
			nrtx_ = 0;
			myrecover_ = 0;
			recv_newack_helper(pkt);
			rtxcount_--;
		} else {
		//partial ack: last_ack_ < ACK < myrecover_
			dupacks_ = 0;
			mydupwnd_ -= (tcph->seqno() - nrtx_ - last_ack_ - 1);
			if(mydupwnd_<0) mydupwnd_=0;
			nrtx_ = 0;
			last_ack_ = tcph->seqno();
			highest_ack_ = last_ack_;
			rtxcount_--;
		}
	} else if (tcph->seqno() == last_ack_) {
		dupacks_++;
		mydupwnd_++;
		if(!myrecover_) {
			myrecover_ = maxseq_;
			if(rtxcount_ <= 0) {
				rtxcount_ = rwnd_; 
			}
			rtx_allow_ = 0;
		}
		if((rtxcount_<=0) && (rtx_allow_==0)) {
			rtx_allow_ = 1;
			if(tcph->ackno() > highnack_) rtx_stop_=highnack_;
			else rtx_stop_ = tcph->ackno();
		} else rtxcount_--;
	}

	nack_ = (int) tcph->ackno();

	if((nack_==rtx_stop_) && ((nack_>rtx_last_) || (nack_==0))) {
		rtx_allow_ = 0;
		rtxcount_ = rwnd_; 
		rtx_stop_ = -1;
	}
	if((nack_ != 0) && (nack_ > highnack_)) {
		highnack_ = nack_;
		if(nack_>highest_ack_) output(nack_, TCP_REASON_NACK);
		rtx_last_ = nack_;
		mydupwnd_--;
		nrtx_++;
		rtxcount_--;
	} else if((nack_ != 0) && (rtx_allow_==1)) {
		if(nack_>highest_ack_) output(nack_, TCP_REASON_NACK);
		rtx_last_ = nack_;
		mydupwnd_--;
		nrtx_++;
	}
	if(rtxcount_ <= 0) {
		rtxcount_=0;
	}
	if (QOption_ && EnblRTTCtr_)
		process_qoption_after_ack (tcph->seqno());
	Packet::free(pkt);
	/*
	 * Try to send more data.
	 */
	send_much(0, 0, maxburst_);
if((nack_==0)&&(rtx_allow_==1)) rtx_last_=-1;
}

/* Print out all the traced variables whenever any one is changed */
void StTcpAgent::traceAll() {
	double curtime;
	Scheduler& s = Scheduler::instance();
	char wrk[500];
	int n;

	curtime = &s ? s.clock() : 0;
	sprintf(wrk,"time: %-8.5f saddr: %-2d sport: %-2d daddr: %-2d dport:"
		" %-2d maxseq: %-4d hiack: %-4d seqno: %-4d cwnd: %-6.3f"
		" ssthresh: %-3d dupacks: %-2d rtt: %-6.3f srtt: %-6.3f"
		" rttvar: %-6.3f bkoff: %-d lastnack: %-4d rwnd: %-4d qwnd: %-6.3f",
		curtime, addr(), port(), daddr(), dport(), int(maxseq_),
		int(highest_ack_), int(t_seqno_), double(cwnd_), int(ssthresh_),
		int(dupacks_), int(t_rtt_)*tcp_tick_,
		(int(t_srtt_) >> T_SRTT_BITS)*tcp_tick_,
		int(t_rttvar_)*tcp_tick_/4.0, int(t_backoff_),
		int(highnack_), int(rwnd_), double(qwnd_));
	n = strlen(wrk);
	wrk[n] = '\n';
	wrk[n+1] = 0;
	if (channel_)
		(void)Tcl_Write(channel_, wrk, n+1);
	wrk[n] = 0;
	return;
}

/* schedule the next burst of data (of size at most maxburst) */
void StTcpAgent::send_helper(int maxburst)
{
	//If there is still data to be sent and there is space in the
	 
	if (t_seqno_ <= highest_ack_ + window() && t_seqno_ < curseq_) {
		if(highest_ack_ + window() - t_seqno_ > maxburst) {
			double interval = tcp_tick_*t_rtt_*maxburst/window();
			if(maxburst >= window()) interval = 0.1;
			burstsnd_timer_.resched(interval);
		}
	}
}


// StTcpSink implementation
NAcker::NAcker() :
	Acker(),
	lastnack_(0)
{
//	printf("New NAcker object!\n");
}

void NAcker::reset() {
	lastnack_ = 0;
	Acker::reset();
}

int NAcker::nack() {
	int nack = next_;
	if(lastnack_ >= nack) nack=lastnack_+1;
	while ((seen_[nack & MWM]) && (nack < maxseen_)) {
		nack++;
	}
	if((nack > lastnack_) && (nack < maxseen_)) {
		lastnack_=nack;
		return(nack);
	} else if(nack==maxseen_) {
		lastnack_=0;
		return(0);
	} else return(0);
}

StTcpSink::StTcpSink(NAcker* acker) :
	DelAckSink(acker),
	acker_(acker),
	rwnd_(1)
{
}


void StTcpSink::ack(Packet* opkt) {
	Packet* npkt = allocpkt();
	// opkt is the "old" packet that was received
	// npkt is the "new" packet being constructed (LACK)
	double now = Scheduler::instance().clock();
	hdr_flags *sf;

	hdr_tcp *otcp = hdr_tcp::access(opkt);
	hdr_tcp *ntcp = hdr_tcp::access(npkt);
	// get the tcp headers
	ntcp->seqno() = acker_->Seqno();
	ntcp->ackno() = acker_->nack();
//	printf("ack seqno\n");

	ntcp->tcp_flags_ = rwnd_;
	// get the cumulative sequence number to put in the ACK; this
	// is just the left edge of the receive window - 1
	ntcp->ts() = now;
	// timestamp the packet

	if (ts_echo_bugfix_)
		ntcp->ts_echo() = acker_->ts_to_echo();
	else
		ntcp->ts_echo() = otcp->ts();
	// echo the original's time stamp

	hdr_ip* oip = hdr_ip::access(opkt);
	hdr_ip* nip = hdr_ip::access(npkt);
	// get the ip headers
	nip->flowid() = oip->flowid();
	// copy the flow id

	hdr_flags* of = hdr_flags::access(opkt);
	hdr_flags* nf = hdr_flags::access(npkt);
	if (save_ != NULL)
		sf = hdr_flags::access(save_);
		// Look at delayed packet being acked.
	if ( (save_ != NULL && sf->cong_action()) || of->cong_action() )
		// Sender has responsed to congestion.
		acker_->update_ecn_unacked(0);
	if ( (save_ != NULL && sf->ect() && sf->ce())  ||
			(of->ect() && of->ce()) )
		// New report of congestion.
		acker_->update_ecn_unacked(1);
	if ( (save_ != NULL && sf->ect()) || of->ect() )
		// Set EcnEcho bit.
		nf->ecnecho() = acker_->ecn_unacked();
	if (!of->ect() && of->ecnecho() ||
		(save_ != NULL && !sf->ect() && sf->ecnecho()) )
		 // This is the negotiation for ECN-capability.
		 // We are not checking for of->cong_action() also.
		 // In this respect, this does not conform to the
		 // specifications in the internet draft
		nf->ecnecho() = 1;
	acker_->append_ack(hdr_cmn::access(npkt),
			   ntcp, otcp->seqno());
	add_to_ack(npkt);

	send(npkt, 0);
}

int StTcpSink::Seqno() {
	return (acker_->Seqno());
}



// SplitTcpAgent implementation


SplitTcpSink::SplitTcpSink(Acker* acker) :
	DelAckSink(acker),
	forwarder_(0)
{
//	printf("New Split TCP sink agent!\n");
	bind("forwarder_", (TclObject **)(&forwarder_));
	bind("useHeaders_", &useHeaders_);
}

int SplitTcpSink::command(int argc, const char*const* argv) {
	if (argc == 3) {
		if (strcmp(argv[1], "forwarder") == 0) {
			forwarder_=(Agent *)TclObject::lookup(argv[2]);
			if (forwarder_ == 0) return TCL_ERROR;
			return TCL_OK;
		}
	}
	return (TcpSink::command(argc, argv));
}

void SplitTcpSink::recvBytes(int nbytes) {
//	if (forwarder_) {
//		if(nbytes > 0) { 
//			if(useHeaders_ == 1) {
//				forwarder_->sendmsg(nbytes-40);
//			} else {
//				forwarder_->sendmsg(nbytes-40);
//			}
//		}
//	}
	if (forwarder_) {
		if(nbytes > 0) { 
			forwarder_->sendmsg(nbytes-40);
		}
	}
}

//
// SplitStTcpAgent implementation
//
//
//
//

SplitStTcpSink::SplitStTcpSink(NAcker* acker) :
	DelAckSink(acker),
	SplitTcpSink(acker),
	StTcpSink(acker)
{
//	printf("New Split St TCP sink agent!\n");
}
